"""
Objects describing the Inference activity, its inputs and outputs as specified
in NIDM-Results.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""
from nidmresults.objects.constants import *
from nidmresults.objects.generic import *
import uuid
from math import erf, sqrt
import rdflib
from prov.model import Literal
from prov.constants import XSD_FLOAT
from prov.model import Identifier


class Inference(object):

    """
    Object representing an Inference step: including an Inference activity, its
    inputs and outputs.
    """

    def __init__(
            self, inference, height_thresh, extent_thresh,
            peak_criteria, cluster_criteria, disp_mask, excursion_set,
            clusters, search_space, software_id):
        super(Inference, self).__init__()
        self.excursion_set = excursion_set
        self.inference_act = inference
        self.height_thresh = height_thresh
        self.extent_thresh = extent_thresh
        self.clusters = clusters
        self.software_id = software_id
        self.peak_criteria = peak_criteria
        self.cluster_criteria = cluster_criteria
        self.disp_mask = disp_mask
        self.search_space = search_space


class InferenceActivity(NIDMObject):

    """
    Object representing an Inference activity.
    """

    def __init__(self, oid=None, tail=None, label=None, contrast_name=None,
                 inference_type=None, partial_degree=None):
        super(InferenceActivity, self).__init__(oid=oid)
        if inference_type is None:
            self.type = NIDM_INFERENCE
        else:
            self.type = inference_type
        self.prov_type = PROV['Activity']
        if tail is None:
            tail = NIDM_ONE_TAILED_TEST
        self.tail = tail
        if label is None:
            label = "Inference"
            if contrast_name:
                label += ": " + contrast_name
        self.label = label
        self.partial_degree = partial_degree

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_Inference: <http://purl.org/nidash/nidm#NIDM_0000049>
prefix nidm_ConjunctionInference: <http://purl.org/nidash/nidm#NIDM_0000011>
prefix nidm_hasAlternativeHypothesis: <http://purl.org/nidash/nidm#NIDM_000009\
7>
prefix spm_PartialConjunctionInference: <http://purl.org/nidash/spm#SPM_000000\
5>
prefix spm_PartialConjunctionDegree: <http://purl.org/nidash/spm#SPM_0000015>

    SELECT DISTINCT * WHERE {
    {
    """ + oid_var + """ a nidm_Inference: .
    } UNION {
    """ + oid_var + """ a nidm_ConjunctionInference: .
    } UNION {
    """ + oid_var + """ a spm_PartialConjunctionInference: .
    }

    """ + oid_var + """ rdfs:label ?label ;
        a ?inference_type ;
        nidm_hasAlternativeHypothesis: ?tail .

    OPTIONAL {""" + oid_var + """ spm_PartialConjunctionDegree: ?partial_degree .} .

    FILTER ( ?inference_type NOT IN (prov:Activity))
        }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """

        # In FSL we have a single thresholding (extent, height) applied to all
        # contrasts
        # FIXME: Deal with two-tailed inference?
        atts = (
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (NIDM_HAS_ALTERNATIVE_HYPOTHESIS, self.tail))

        if self.partial_degree is not None:
            atts += (
                (SPM_PARTIAL_CONJUNCTION_DEGREE, self.partial_degree),)

        self.add_attributes(atts)


class ExcursionSet(NIDMObject):

    """
    Object representing a ExcursionSet entity.
    """

    def __init__(self, location, coord_space, visu=None,
                 oid=None, fmt=None, label=None,
                 sha=None, filename=None, inference=None, suffix='',
                 clust_map=None, mip=None, num_clusters=None, p_value=None):
        super(ExcursionSet, self).__init__(oid)
        if not filename:
            filename = 'ExcursionSet' + suffix + '.nii.gz'
        else:
            filename = location
        self.filename = filename
        self.file = NIDMFile(self.id, location, filename, sha)
        self.type = NIDM_EXCURSION_SET_MAP
        self.prov_type = PROV['Entity']
        self.visu = visu
        if label is None:
            label = "Excursion Set Map"
        self.label = label
        self.coord_space = coord_space
        self.clust_map = clust_map
        self.mip = mip
        # FIXME Not used for export yet (only for reading)
        self.inference = inference
        self.num_clusters = num_clusters
        self.p_value = p_value

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ExcursionSetMap: <http://purl.org/nidash/nidm#NIDM_0000025>
prefix nidm_hasClusterLabelsMap: <http://purl.org/nidash/nidm#NIDM_0000098>
prefix nidm_hasMaximumIntensityProjection: <http://purl.org/nidash/nidm#NIDM_0\
000138>
prefix nidm_inCoordinateSpace: <http://purl.org/nidash/nidm#NIDM_0000104>
prefix nidm_numberOfSupraThresholdClusters: <http://purl.org/nidash/nidm#NIDM_\
0000111>
prefix nidm_pValue: <http://purl.org/nidash/nidm#NIDM_0000114>

SELECT DISTINCT * WHERE {

""" + oid_var + """ a nidm_ExcursionSetMap: ;
    prov:atLocation ?location ;
    rdfs:label ?label ;
    dct:format ?fmt ;
    nfo:fileName ?filename ;
    crypto:sha512 ?sha .

    OPTIONAL {""" + oid_var + """ nidm_numberOfSupraThresholdClusters: ?num_clusters .} .
    OPTIONAL {""" + oid_var + """ nidm_pValue: ?p_value .} .
}

ORDER BY ?peak_label
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        # Create "Excursion set" entity
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (PROV['label'], self.label),
        ))

        if self.visu is not None:
            self.add_attributes((
                (DC['description'], self.visu.id),
            ))

        if self.clust_map is not None:
            self.add_attributes((
                (NIDM_HAS_CLUSTER_LABELS_MAP, self.clust_map.id),
            ))

        if self.mip is not None:
            self.add_attributes((
                (NIDM_HAS_MAXIMUM_INTENSITY_PROJECTION, self.mip.id),
            ))

        if self.num_clusters is not None:
            self.add_attributes((
                (NIDM_NUMBER_OF_CLUSTERS, self.num_clusters),
            ))

        if self.p_value is not None:
            self.add_attributes((
                (NIDM_P_VALUE, self.p_value),
            ))


class ClusterLabelsMap(NIDMObject):

    """
    Object representing a ClusterLabelsMap entity.
    """

    def __init__(self, location, coord_space,
                 oid=None, fmt=None, label=None,
                 sha=None, filename=None, suffix='', temporary=False):
        super(ClusterLabelsMap, self).__init__(oid)
        if not filename:
            filename = 'ClusterLabels' + suffix + '.nii.gz'
        self.filename = filename
        self.file = NIDMFile(self.id, location, filename, sha,
                             temporary=temporary)
        self.type = NIDM_CLUSTER_LABELS_MAP
        self.prov_type = PROV['Entity']
        if label is None:
            label = "Cluster Labels Map"
        self.label = label
        self.coord_space = coord_space

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ClusterLabelsMap: <http://purl.org/nidash/nidm#NIDM_0000008>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ClusterLabelsMap: ;
        nfo:fileName ?filename ;
        crypto:sha512 ?sha ;
        prov:atLocation ?location ;
        dct:format ?fmt .
        }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        # Create "Cluster Labels Map" entity
        self.add_attributes((
            (PROV['type'], self.type),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (PROV['label'], self.label)
        ))


class HeightThreshold(NIDMObject):

    """
    Object representing a HeightThreshold entity.
    """

    def __init__(self, stat_threshold=None, p_corr_threshold=None,
                 p_uncorr_threshold=None, threshold_type=None, value=None,
                 label=None, version={'num': '1.3.0'}, oid=None,
                 equiv_thresh=None):
        super(HeightThreshold, self).__init__(oid=oid)
        if not stat_threshold and not p_corr_threshold and \
           not p_uncorr_threshold and not value:
            raise Exception('No threshold defined')

        if isinstance(threshold_type, str):
            threshold_type = Identifier(threshold_type)

        thresh_desc = ""
        if stat_threshold is not None:
            thresh_desc = "Z>" + str(stat_threshold)
            if version['num'] == "1.0.0":
                user_threshold_type = "Z-Statistic"
            else:
                threshold_type = OBO_STATISTIC
                value = stat_threshold
        elif p_uncorr_threshold is not None:
            thresh_desc = "p<" + \
                str(p_uncorr_threshold) + " (uncorrected)"
            if version['num'] == "1.0.0":
                user_threshold_type = "p-value uncorrected"
            else:
                threshold_type = NIDM_P_VALUE_UNCORRECTED_CLASS
                value = p_uncorr_threshold
        elif p_corr_threshold is not None:
            thresh_desc = "p<" + str(p_corr_threshold) + " (FWE)"
            if version['num'] == "1.0.0":
                user_threshold_type = "p-value FWE"
            else:
                threshold_type = OBO_P_VALUE_FWER
                value = p_corr_threshold

        if version['num'] == "1.0.0":
            self.user_threshold_type = user_threshold_type
            self.p_uncorr_threshold = p_uncorr_threshold
            self.p_corr_threshold = p_corr_threshold
            self.stat_threshold = stat_threshold
        else:
            self.value = value
            self.threshold_type = threshold_type

        if not label:
            self.label = "Height Threshold: " + thresh_desc
        else:
            self.label = label

        self.type = NIDM_HEIGHT_THRESHOLD
        self.prov_type = PROV['Entity']
        self.equiv_thresh = equiv_thresh

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_HeightThreshold: <http://purl.org/nidash/nidm#NIDM_0000034>
prefix nidm_hasAlternativeHypothesis: <http://purl.org/nidash/nidm#NIDM_000009\
7>

    SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_HeightThreshold: ;
        a ?threshold_type ;
        rdfs:label ?label ;
        prov:value ?value .

    FILTER ( ?threshold_type NOT IN (prov:Entity, nidm_HeightThreshold:) )
}
        """
        return query

    def export(self, version, export_dir):
        """
        Create prov entities and activities.
        """

        atts = [
            (PROV['type'], self.type),
            (PROV['label'], self.label),
        ]

        if version['num'] == "1.0.0":
            atts += [
                (NIDM_USER_SPECIFIED_THRESHOLD_TYPE, self.user_threshold_type),
                (PROV['value'], self.stat_threshold),
                (NIDM_P_VALUE_UNCORRECTED, self.p_uncorr_threshold),
                (NIDM_P_VALUE_FWER, self.p_corr_threshold)
                ]
        else:
            atts += [
                (PROV['type'], self.threshold_type),
                (PROV['value'], self.value)
                ]

        if self.equiv_thresh is not None:
            for equiv in self.equiv_thresh:
                atts += [
                    (NIDM_EQUIVALENT_THRESHOLD, equiv.id)
                ]

        self.add_attributes([(k, v) for k, v in atts if v is not None])


class ExtentThreshold(NIDMObject):

    """
    Object representing an ExtentThreshold entity.
    """

    def __init__(self, extent=None, p_corr=None, p_uncorr=None,
                 extent_rsl=None, label=None, version={'num': '1.3.0'},
                 value=None, oid=None, equiv_thresh=None, threshold_type=None):
        super(ExtentThreshold, self).__init__(oid=oid)
        self.type = NIDM_EXTENT_THRESHOLD
        self.prov_type = PROV['Entity']

        thresh_desc = ""

        if threshold_type is not None:
            self.threshold_type = threshold_type
        else:
            if extent is not None:
                thresh_desc = "k>" + str(extent)
                # NIDM-Results 1.0.0
                user_threshold_type = "Cluster-size in voxels"
                # NIDM-Results > 1.0.0
                threshold_type = OBO_STATISTIC
            elif p_uncorr is not None:
                thresh_desc = "p<" + str(self.p_uncorr) + " (uncorrected)"
                # NIDM-Results 1.0.0
                user_threshold_type = "p-value uncorrected"
                # NIDM-Results > 1.0.0
                threshold_type = NIDM_P_VALUE_UNCORRECTED_CLASS
                value = p_uncorr
            elif p_corr is not None:
                thresh_desc = "p<" + str(p_corr) + " (FWE)"
                # NIDM-Results 1.0.0
                user_threshold_type = "p-value FWE"
                # NIDM-Results > 1.0.0
                threshold_type = OBO_P_VALUE_FWER
                value = p_corr
            else:
                thresh_desc = "k>=0"
                extent = 0
                if version['num'] == "1.0.0":
                    p_uncorr = 1.0
                    p_corr = 1.0
                    user_threshold_type = None
                else:
                    threshold_type = OBO_STATISTIC

        self.threshold_type = threshold_type
        self.value = value

        if version['num'] == "1.0.0":
            self.user_threshold_type = user_threshold_type
            self.p_uncorr = p_uncorr
            self.p_corr = p_corr
        else:
            self.threshold_type = threshold_type

        self.extent = extent
        self.extent_rsl = extent_rsl

        if label is None:
            self.label = "Extent Threshold: " + thresh_desc
        else:
            self.label = label

        self.equiv_thresh = equiv_thresh

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ExtentThreshold: <http://purl.org/nidash/nidm#NIDM_0000026>
prefix nidm_clusterSizeInVoxels: <http://purl.org/nidash/nidm#NIDM_0000084>
prefix nidm_clusterSizeInResels: <http://purl.org/nidash/nidm#NIDM_0000156>

    SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ExtentThreshold: ;
        a ?threshold_type ;
        rdfs:label ?label  .
    OPTIONAL {""" + oid_var + """ prov:value ?value .} .
    OPTIONAL {""" + oid_var + """ nidm_clusterSizeInVoxels: ?extent .} .
    OPTIONAL {""" + oid_var + """ nidm_clusterSizeInResels: ?extent_rsl .} .

    FILTER ( ?threshold_type NOT IN (prov:Entity, nidm_ExtentThreshold:) )
}
        """
        return query

    def export(self, version, export_dir):
        """
        Create prov entities and activities.
        """
        atts = [
            (PROV['type'], self.type),
        ]

        atts += [
            (PROV['label'], self.label)
        ]

        if self.extent_rsl is not None:
            atts += [
                (NIDM_CLUSTER_SIZE_IN_RESELS, self.extent_rsl),
            ]

        if self.extent is not None:
            atts += [
                (NIDM_CLUSTER_SIZE_IN_VOXELS, self.extent),
            ]

        if version['num'] == "1.0.0":
            atts += [
                (NIDM_USER_SPECIFIED_THRESHOLD_TYPE, self.user_threshold_type),
                (NIDM_P_VALUE_UNCORRECTED, self.p_uncorr),
                (NIDM_P_VALUE_FWER, self.p_corr)
            ]
        else:
            atts += [
                (PROV['type'], self.threshold_type)
            ]
            if self.value is not None:
                atts += [
                    (PROV['value'], self.value)
                ]

        if self.equiv_thresh is not None:
            for equiv in self.equiv_thresh:
                atts += [
                    (NIDM_EQUIVALENT_THRESHOLD, equiv.id)
                ]

        self.add_attributes([(k, v) for k, v in atts if v is not None])


class Cluster(NIDMObject):

    """
    Object representing a Cluster entity.
    """

    def __init__(self, cluster_num, size, pFWER, peaks,
                 x=None, y=None, z=None, x_std=None, y_std=None, z_std=None,
                 suffix='', clust_size_resels=None, pFDR=None, punc=None,
                 label=None, oid=None, cog=None):
        super(Cluster, self).__init__(oid=oid)
        self.num = cluster_num
        if cog is not None:
            self.cog = cog
        else:
            if x and y and z:
                self.cog = CenterOfGravity(
                    cluster_num, x=x, y=y, z=z, x_std=x_std, y_std=y_std,
                    z_std=z_std)
            else:
                self.cog = None
        self.peaks = peaks
        self.size = size
        self.pFWER = pFWER
        self.type = NIDM_SIGNIFICANT_CLUSTER
        self.prov_type = PROV['Entity']
        self.punc = punc
        self.pFDR = pFDR
        if not label:
            cluster_naming = "Supra-Threshold Cluster"
            self.label = "%s %04d" % (cluster_naming, self.num)
        else:
            self.label = label

        self.clust_size_resels = clust_size_resels

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_SupraThresholdCluster: <http://purl.org/nidash/nidm#NIDM_0000070>
prefix nidm_clusterSizeInVoxels: <http://purl.org/nidash/nidm#NIDM_0000084>
prefix nidm_clusterLabelId: <http://purl.org/nidash/nidm#NIDM_0000082>
prefix nidm_clusterSizeInResels: <http://purl.org/nidash/nidm#NIDM_0000156>
prefix nidm_pValueUncorrected: <http://purl.org/nidash/nidm#NIDM_0000116>
prefix nidm_pValueFWER: <http://purl.org/nidash/nidm#NIDM_0000115>
prefix nidm_qValueFDR: <http://purl.org/nidash/nidm#NIDM_0000119>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_SupraThresholdCluster: ;
        rdfs:label ?label ;
        nidm_clusterSizeInVoxels: ?size ;
        nidm_clusterLabelId: ?cluster_num .

    OPTIONAL {""" + oid_var + """ nidm_clusterSizeInResels: ?clust_size_resels .} .
    OPTIONAL {""" + oid_var + """ nidm_pValueUncorrected: ?punc .} .
    OPTIONAL {""" + oid_var + """ nidm_pValueFWER: ?pFWER .} .
    OPTIONAL {""" + oid_var + """ nidm_qValueFDR: ?pFDR .} .
}
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        if nidm_version['num'] in ["1.0.0", "1.1.0"]:
            self.label = self.label.replace("Supra-Threshold", "Significant")

        # FIXME deal with multiple contrasts
        atts = (
            (PROV['type'], NIDM_SIGNIFICANT_CLUSTER),
            (PROV['label'], self.label),
            (NIDM_CLUSTER_LABEL_ID, self.num),
            (NIDM_CLUSTER_SIZE_IN_VOXELS, self.size)
            )

        if self.clust_size_resels is not None:
            atts = atts + (
                (NIDM_CLUSTER_SIZE_IN_RESELS, self.clust_size_resels),
                )

        if self.punc is not None:
            atts = atts + (
                (NIDM_P_VALUE_UNCORRECTED,
                    Literal(self.punc, datatype=XSD_FLOAT)),
                )

        if self.pFDR is not None:
            atts = atts + (
                (NIDM_Q_VALUE_FDR, Literal(self.pFDR, datatype=XSD_FLOAT)),
                )

        if self.pFWER is not None:
            atts = atts + (
                (NIDM_P_VALUE_FWER, Literal(self.pFWER, datatype=XSD_FLOAT)),
                )

        self.add_attributes(atts)


class DisplayMaskMap(NIDMObject):

    """
    Object representing a DisplayMaskMap entity.
    """
    def __init__(self, contrast_num, mask_file, mask_num, coord_space,
                 sha=None, filename=None, fmt=None, label=None, oid=None,
                 derfrom_id=None, derfrom_filename=None, derfrom_fmt=None,
                 derfrom_sha=None, isderfrommap=False):
        super(DisplayMaskMap, self).__init__(oid=oid)
        if not filename:
            filename = 'DisplayMask' + str(mask_num) + '.nii.gz'
        self.file = NIDMFile(self.id, mask_file, filename,
                             sha=sha, fmt=fmt)
        self.coord_space = coord_space
        self.type = NIDM_DISPLAY_MASK_MAP
        self.prov_type = PROV['Entity']
        if not label:
            self.label = "Display Mask Map " + str(mask_num)
        else:
            self.label = label
        if derfrom_id is not None:
            self.derfrom = DisplayMaskMap(
                None, None, None,
                coord_space=None, oid=derfrom_id,
                filename=derfrom_filename, sha=derfrom_sha,
                fmt=derfrom_fmt,
                isderfrommap=True)
        else:
            self.derfrom = None
        self.isderfrommap = isderfrommap

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_DisplayMaskMap: <http://purl.org/nidash/nidm#NIDM_0000020>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_DisplayMaskMap: ;
        rdfs:label ?label ;
        nfo:fileName ?filename ;
        crypto:sha512 ?sha ;
        prov:atLocation ?mask_file ;
        dct:format ?fmt .

    OPTIONAL {""" + oid_var + """ prov:wasDerivedFrom ?derfrom_id .

    ?derfrom_id a nidm_DisplayMaskMap: ;
        nfo:fileName ?derfrom_filename ;
        dct:format ?derfrom_fmt ;
        crypto:sha512 ?derfrom_sha .
     } .

}
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        atts = (
            (PROV['type'], self.type),
            )

        if not self.isderfrommap:
            atts = atts + (
                (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
                (PROV['label'], self.label))

        self.add_attributes(atts)


class PeakCriteria(NIDMObject):

    """
    Object representing a PeakCriteria entity.
    """

    def __init__(self, contrast_num, peak_dist, num_peak=None, label=None,
                 oid=None):
        super(PeakCriteria, self).__init__(oid=oid)
        self.num_peak = num_peak
        self.peak_dist = peak_dist
        self.type = NIDM_PEAK_DEFINITION_CRITERIA
        self.prov_type = PROV['Entity']
        if not label:
            self.label = "Peak Definition Criteria"
        else:
            self.label = label

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_PeakDefinitionCriteria: <http://purl.org/nidash/nidm#NIDM_0000063>
prefix nidm_minDistanceBetweenPeaks: <http://purl.org/nidash/nidm#NIDM_0000109>
prefix nidm_maxNumberOfPeaksPerCluster: <http://purl.org/nidash/nidm#NIDM_0000\
108>

    SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_PeakDefinitionCriteria: ;
        rdfs:label ?label ;
        nidm_minDistanceBetweenPeaks: ?peak_dist .

    OPTIONAL { """ + oid_var + """ nidm_maxNumberOfPeaksPerCluster: ?num_peak .} .
}
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        num_peak = ()
        if self.num_peak:
            num_peak = ((NIDM_MAX_NUMBER_OF_PEAKS_PER_CLUSTER, self.num_peak),)

        # Create "Peak definition criteria" entity
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (NIDM_MIN_DISTANCE_BETWEEN_PEAKS, self.peak_dist)
            ) + num_peak)


class ClusterCriteria(NIDMObject):

    """
    Object representing a ClusterCriteria entity.
    """

    def __init__(self, contrast_num, connectivity, label=None, oid=None):
        super(ClusterCriteria, self).__init__(oid=oid)
        self.connectivity = connectivity
        self.type = NIDM_CLUSTER_DEFINITION_CRITERIA
        self.prov_type = PROV['Entity']
        if not label:
            self.label = ("Cluster Connectivity Criterion: " +
                          str(self.connectivity))
        else:
            self.label = label

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ClusterDefinitionCriteria: <http://purl.org/nidash/nidm#NIDM_00000\
07>
prefix nidm_hasConnectivityCriterion: <http://purl.org/nidash/nidm#NIDM_000009\
9>

    SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ClusterDefinitionCriteria: ;
        rdfs:label ?label ;
        nidm_hasConnectivityCriterion: ?connectivity .
}
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        # Create "Cluster definition criteria" entity
        if isinstance(self.connectivity, int):
            if self.connectivity == 6:
                self.connectivity = NIDM_VOXEL6CONNECTED
            elif self.connectivity == 18:
                self.connectivity = NIDM_VOXEL18CONNECTED
            elif self.connectivity == 26:
                self.connectivity = NIDM_VOXEL26CONNECTED

        # FIXME if connectivity is missing
        if self.connectivity is not None:
            atts = (
                (PROV['type'], self.type),
                (PROV['label'], self.label),
                (NIDM_HAS_CONNECTIVITY_CRITERION, self.connectivity))
        else:
            atts = (
                (PROV['type'], NIDM_CLUSTER_DEFINITION_CRITERIA),
                (PROV['label'], label))

        self.add_attributes(atts)


class CenterOfGravity(NIDMObject):

    """
    Object representing a CenterOfGravity entity.
    """

    def __init__(self, cluster_num, x=None, y=None, z=None, x_std=None,
                 y_std=None, z_std=None, oid=None, coord_vector=None,
                 coord_vector_std=None, label=None, coord_id=None):
        # Note: coord_id argument is only here for compatibility
        # with the query outputs

        super(CenterOfGravity, self).__init__(oid=oid)
        self.cluster_num = cluster_num
        self.coordinate = Coordinate("%04d" % cluster_num, x=x, y=y, z=z,
                                     x_std=x_std, y_std=y_std, z_std=z_std,
                                     coord_vector_std=coord_vector_std,
                                     coord_vector=coord_vector, oid=coord_id)
        self.type = NIDM_CLUSTER_CENTER_OF_GRAVITY
        self.prov_type = PROV['Entity']
        if label is None:
            label = "Center of gravity " + str(self.cluster_num)
        self.label = label

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_ClusterCenterOfGravity: <http://purl.org/nidash/nidm#NIDM_0000140>
prefix nidm_coordinateVector: <http://purl.org/nidash/nidm#NIDM_0000086>
prefix nidm_coordinateVectorInVoxels: <http://purl.org/nidash/nidm#NIDM_000013\
9>

    SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_ClusterCenterOfGravity: ;
        rdfs:label ?label ;
        prov:atLocation ?coord_id .

    ?coord_id a nidm_Coordinate: ;
        nidm_coordinateVector: ?coord_vector_std .

    OPTIONAL { ?coord_id nidm_coordinateVectorInVoxels: ?coord_vector .} .
}
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (PROV['location'], self.coordinate.id)))


class SearchSpace(NIDMObject):

    """
    Object representing a SearchSpace entity.
    """

    def __init__(self, search_space_file, vol_in_voxels, vol_in_units,
                 vol_in_resels, resel_size_in_voxels,
                 random_field_stationarity, noise_fwhm_in_voxels,
                 noise_fwhm_in_units, coord_space,
                 expected_num_voxels=None, expected_num_clusters=None,
                 height_critical_fwe05=None, height_critical_fdr05=None,
                 extent_critical_fwe05=None, extent_critical_fdr05=None,
                 search_vol_geom=None, noise_roughness=None,
                 filename=None, sha=None, fmt=None,
                 label=None, oid=None):
        super(SearchSpace, self).__init__(oid=oid)
        if not filename:
            filename = 'SearchSpaceMask.nii.gz'
        self.file = NIDMFile(self.id, search_space_file, filename,
                             sha=sha, fmt=fmt)
        self.coord_space = coord_space
        self.resel_size_in_voxels = resel_size_in_voxels
        self.search_volume_in_voxels = vol_in_voxels
        self.search_volume_in_units = vol_in_units
        self.search_volume_in_resels = vol_in_resels
        self.rf_stationarity = random_field_stationarity
        self.noise_fwhm_in_voxels = noise_fwhm_in_voxels
        self.noise_fwhm_in_units = noise_fwhm_in_units
        self.type = NIDM_SEARCH_SPACE_MASK_MAP
        self.prov_type = PROV['Entity']
        self.label = "Search Space Mask Map"
        self.expected_num_voxels = expected_num_voxels
        self.expected_num_clusters = expected_num_clusters
        self.height_critical_fwe05 = height_critical_fwe05
        self.height_critical_fdr05 = height_critical_fdr05
        self.extent_critical_fwe05 = extent_critical_fwe05
        self.extent_critical_fdr05 = extent_critical_fdr05
        self.search_vol_geom = search_vol_geom
        self.noise_roughness = noise_roughness

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_SearchSpaceMaskMap: <http://purl.org/nidash/nidm#NIDM_0000068>
prefix nidm_expectedNumberOfVoxelsPerCluster: <http://purl.org/nidash/nidm#NID\
M_0000143>
prefix nidm_expectedNumberOfClusters: <http://purl.org/nidash/nidm#NIDM_000014\
1>
prefix nidm_heightCriticalThresholdFWE05: <http://purl.org/nidash/nidm#NIDM_00\
00147>
prefix nidm_heightCriticalThresholdFDR05: <http://purl.org/nidash/nidm#NIDM_00\
00146>
prefix nidm_searchVolumeInVoxels: <http://purl.org/nidash/nidm#NIDM_0000121>
prefix nidm_searchVolumeInUnits: <http://purl.org/nidash/nidm#NIDM_0000136>
prefix nidm_searchVolumeInResels: <http://purl.org/nidash/nidm#NIDM_0000149>
prefix nidm_reselSizeInVoxels: <http://purl.org/nidash/nidm#NIDM_0000148>
prefix nidm_noiseFWHMInVoxels: <http://purl.org/nidash/nidm#NIDM_0000159>
prefix nidm_noiseFWHMInUnits: <http://purl.org/nidash/nidm#NIDM_0000157>
prefix nidm_randomFieldStationarity: <http://purl.org/nidash/nidm#NIDM_0000120>
prefix spm_smallestSignificantClusterSizeInVoxelsFWE05: <http://purl.org/nidas\
h/spm#SPM_0000014>
prefix spm_smallestSignificantClusterSizeInVoxelsFDR05: <http://purl.org/nidas\
h/spm#SPM_0000013>
prefix spm_searchVolumeReselsGeometry: <http://purl.org/nidash/spm#SPM_0000010>
prefix nidm_noiseRoughnessInVoxels: <http://purl.org/nidash/nidm#NIDM_0000145>

SELECT DISTINCT * WHERE {

""" + oid_var + """ a nidm_SearchSpaceMaskMap: ;
    rdfs:label ?label ;
    nidm_searchVolumeInVoxels: ?vol_in_voxels ;
    nidm_searchVolumeInUnits: ?vol_in_units ;
    nidm_searchVolumeInResels: ?vol_in_resels ;
    nidm_reselSizeInVoxels: ?resel_size_in_voxels ;
    nidm_reselSizeInVoxels: ?resel_size_in_voxels ;
    nidm_noiseFWHMInVoxels: ?noise_fwhm_in_voxels ;
    nidm_noiseFWHMInUnits: ?noise_fwhm_in_units ;
    nidm_randomFieldStationarity: ?random_field_stationarity ;
    prov:atLocation ?search_space_file ;
    dct:format ?fmt ;
    nfo:fileName ?filename ;
    crypto:sha512 ?sha .

    OPTIONAL {""" + oid_var + """ nidm_expectedNumberOfVoxelsPerCluster: ?expected_num_voxels } .
    OPTIONAL {""" + oid_var + """ nidm_expectedNumberOfClusters: ?expected_num_clusters } .
    OPTIONAL {""" + oid_var + """ nidm_heightCriticalThresholdFWE05: ?height_critical_fwe05 } .
    OPTIONAL {""" + oid_var + """ nidm_heightCriticalThresholdFDR05: ?height_critical_fdr05 } .
    OPTIONAL {""" + oid_var + """ spm_smallestSignificantClusterSizeInVoxelsFWE05: ?extent_critical_fwe05 } .
    OPTIONAL {""" + oid_var + """ spm_smallestSignificantClusterSizeInVoxelsFDR05: ?extent_critical_fdr05 } .
    OPTIONAL {""" + oid_var + """ spm_searchVolumeReselsGeometry: ?search_vol_geom } .
    OPTIONAL {""" + oid_var + """ nidm_noiseRoughnessInVoxels: ?noise_roughness } .

}

        """
        return query

    # Generate prov for search space entity generated by the inference activity
    def export(self, version, export_dir):
        """
        Create prov entities and activities.
        """
        atts = (
            (PROV['label'], self.label),
            (PROV['type'], NIDM_SEARCH_SPACE_MASK_MAP),
            (NIDM_RANDOM_FIELD_STATIONARITY, self.rf_stationarity),
            (NIDM_IN_COORDINATE_SPACE, self.coord_space.id),
            (NIDM_SEARCH_VOLUME_IN_VOXELS, self.search_volume_in_voxels),
            (NIDM_SEARCH_VOLUME_IN_UNITS, self.search_volume_in_units),
            (NIDM_SEARCH_VOLUME_IN_RESELS, self.search_volume_in_resels),
            (NIDM_RESEL_SIZE_IN_VOXELS, self.resel_size_in_voxels))

        # Noise FWHM was introduced in NIDM-Results 1.1.0
        if self.noise_fwhm_in_voxels is not None:
            if (version['major'] > 1) or \
               (version['major'] >= 1 and
                    (version['minor'] > 0 or version['revision'] > 0)):
                atts = atts + (
                    (NIDM_NOISE_FWHM_IN_VOXELS, self.noise_fwhm_in_voxels),
                    (NIDM_NOISE_FWHM_IN_UNITS, self.noise_fwhm_in_units))

        if self.expected_num_voxels is not None:
            atts = atts + ((NIDM_EXPECTED_NUMBER_OF_VOXELS_PER_CLUSTER,
                            self.expected_num_voxels),)

        if self.expected_num_clusters is not None:
            atts = atts + ((NIDM_EXPECTED_NUMBER_OF_CLUSTERS,
                            self.expected_num_clusters),)

        if self.height_critical_fwe05 is not None:
            atts = atts + ((NIDM_HEIGHT_CRITICAL_THRESHOLD_FWE_05,
                            self.height_critical_fwe05),)

        if self.height_critical_fdr05 is not None:
            atts = atts + ((NIDM_HEIGHT_CRITICAL_THRESHOLD_FDR_05,
                            self.height_critical_fdr05),)

        if self.extent_critical_fwe05 is not None:
            atts = atts + ((
                SPM_SMALLEST_SIGNIFICANT_CLUSTER_SIZE_IN_VOXELS_FWE05,
                self.extent_critical_fwe05),)

        if self.extent_critical_fdr05 is not None:
            atts = atts + ((
                SPM_SMALLEST_SIGNIFICANT_CLUSTER_SIZE_IN_VOXELS_FDR05,
                self.extent_critical_fdr05),)

        if self.search_vol_geom is not None:
            atts = atts + ((SPM_SEARCH_VOLUME_RESELS_GEOMETRY,
                            self.search_vol_geom),)

        if self.noise_roughness:
            atts = atts + ((NIDM_NOISE_ROUGHNESS_IN_VOXELS,
                            self.noise_roughness),)

        # Create "Search Space Mask map" entity
        self.add_attributes(atts)


class Coordinate(NIDMObject):

    """
    Object representing a Coordinate entity.
    """

    def __init__(self, label_id, coord_vector=None, coord_vector_std=None,
                 x=None, y=None, z=None, x_std=None, y_std=None, z_std=None,
                 label=None, oid=None):

        super(Coordinate, self).__init__(oid=oid)
        self.label_id = label_id
        if x is not None and y is not None and z is not None:
            self.coord_vector = [x, y, z]
        else:
            if coord_vector and not type(coord_vector) is list:
                coord_vector = json.loads(coord_vector)
            self.coord_vector = coord_vector
        if x_std is not None and y_std is not None and z_std is not None:
            self.coord_vector_std = [x_std, y_std, z_std]
        else:
            if coord_vector_std and not type(coord_vector_std) is list:
                coord_vector_std = json.loads(coord_vector_std)
            self.coord_vector_std = coord_vector_std

        self.type = NIDM_COORDINATE
        self.prov_type = PROV['Entity']
        if label is not None:
            self.label = label
        else:
            self.label = "Coordinate " + self.label_id

    def __str__(self):
        return '%s\t%s' % (self.label, self.coord_vector)

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        # We can not have this as a dictionnary because we want to keep the
        # duplicate prov:type attribute
        atts = (  # (PROV['type'],PROV['Location']),
            (PROV['type'], NIDM_COORDINATE),
            (PROV['type'], PROV['Location']),
            (PROV['label'], self.label)
            )

        if self.coord_vector is not None:
            atts = atts +\
                ((NIDM_COORDINATE_VECTOR_IN_VOXELS,
                  json.dumps(self.coord_vector)),)

        # FSL unnormalised subject-level analyses do not provide coordinates in
        # voxels
        if self.coord_vector_std is not None:
            atts = atts +\
                ((NIDM_COORDINATE_VECTOR, json.dumps(self.coord_vector_std)),)

        self.add_attributes(atts)


class Peak(NIDMObject):

    """
    Object representing a Peak entity.
    """

    def __init__(self, equiv_z, p_unc=None, p_fwer=None, label=None,
                 coord_label=None, exc_set_id=None, oid=None, suffix='',
                 p_fdr=None, value=None, coord_id=None, *args, **kwargs):
        super(Peak, self).__init__(oid)
        # FIXME: Currently assumes less than 10 clusters per contrast
        # cluster_num = cluster_index
        # FIXME: Currently assumes less than 100 peaks
        if oid is not None:
            self.label = label
            peak_unique_id = label[5:]
            peak_index = peak_unique_id
            # cluster_index, peak_index = peak_unique_id.split("_")
        else:
            peak_unique_id = suffix
            self.label = "Peak " + peak_unique_id
        self.equiv_z = equiv_z
        self.p_unc = p_unc
        self.p_fwer = p_fwer
        self.coordinate = Coordinate(
            str(peak_unique_id), label=coord_label, oid=coord_id, **kwargs)
        self.type = NIDM_PEAK
        self.prov_type = PROV['Entity']
        # self.cluster = cluster_id
        self.exc_set_id = exc_set_id
        self.value = value
        self.p_fdr = p_fdr

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_Peak: <http://purl.org/nidash/nidm#NIDM_0000062>
prefix nidm_pValueUncorrected: <http://purl.org/nidash/nidm#NIDM_0000116>
prefix nidm_equivalentZStatistic: <http://purl.org/nidash/nidm#NIDM_0000092>
prefix nidm_pValueFWER: <http://purl.org/nidash/nidm#NIDM_0000115>
prefix nidm_qValueFDR: <http://purl.org/nidash/nidm#NIDM_0000119>
prefix nidm_coordinateVectorInVoxels: <http://purl.org/nidash/nidm#NIDM_000013\
9>
prefix nidm_coordinateVector: <http://purl.org/nidash/nidm#NIDM_0000086>

SELECT DISTINCT * WHERE {
    """ + oid_var + """ a nidm_Peak: ;
        rdfs:label ?label ;
        prov:atLocation ?coord_id .

    ?coord_id a nidm_Coordinate: ;
        rdfs:label ?coord_label ;
        nidm_coordinateVector: ?coord_vector_std .

    OPTIONAL {?coord_id nidm_coordinateVectorInVoxels: ?coord_vector .} .

    OPTIONAL {""" + oid_var + """ prov:value ?value .} .
    OPTIONAL {""" + oid_var + """ nidm_pValueUncorrected: ?p_unc .} .
    OPTIONAL {""" + oid_var + """ nidm_equivalentZStatistic: ?equiv_z .} .
    OPTIONAL {""" + oid_var + """ nidm_pValueFWER: ?p_fwer .} .
    OPTIONAL {""" + oid_var + """ nidm_qValueFDR: ?p_fdr .} .
        }
        """
        return query

    def __str__(self):
        return '%s \tz=%.2f \tp=%.2e (unc.) \t%s' % (
            self.label, self.equiv_z, self.p_unc, str(self.coordinate))

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        if self.p_unc is None:
            norm_cdf_z = (1.0 + erf(self.equiv_z / sqrt(2.0))) / 2.0
            self.p_unc = 1 - norm_cdf_z

        atts = (
            (PROV['type'], self.type),
            (PROV['label'], self.label),
            (PROV['location'], self.coordinate.id))

        if self.value is not None:
            atts = atts + (
                (PROV['value'], self.value),
                )

        if self.p_unc is not None:
            atts = atts + (
                (NIDM_P_VALUE_UNCORRECTED,
                 Literal(self.p_unc, datatype=XSD_FLOAT)),
                )

        if self.equiv_z is not None:
            atts = atts + (
                (NIDM_EQUIVALENT_ZSTATISTIC,
                 Literal(self.equiv_z, datatype=XSD_FLOAT)),
                )

        if self.p_fdr is not None:
            atts = atts + (
                (NIDM_Q_VALUE_FDR,
                 Literal(self.p_fdr, datatype=XSD_FLOAT)),
                )

        if self.p_fwer is not None:
            atts = atts + (
                (NIDM_P_VALUE_FWER,
                 Literal(self.p_fwer, datatype=XSD_FLOAT)),
                )

        self.add_attributes(atts)



"""
Export neuroimaging results created by neuroimaging software packages
(FSL, AFNI, ...) following NIDM-Results specification.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""


from prov.model import ProvBundle, ProvDocument
import os
import datetime
from nidmresults.objects.constants import *
from nidmresults.objects.modelfitting import *
from nidmresults.objects.contrast import *
from nidmresults.objects.inference import *
from io import open
import uuid
import csv
import tempfile
import zipfile
from builtins import input


class NIDMExporter():

    """
    Generic class to parse a result directory to extract the pieces of
    information to be stored in NIDM-Results and to generate a NIDM-Results
    export.
    """

    def __init__(self, version, out_dir, zipped=True):
        out_dirname = os.path.basename(out_dir)
        out_path = os.path.dirname(out_dir)

        # Create output path from output name
        self.zipped = zipped
        if not self.zipped:
            out_dirname = out_dirname+".nidm"
        else:
            out_dirname = out_dirname+".nidm.zip"
        out_dir = os.path.join(out_path, out_dirname)

        # Quit if output path already exists and user doesn't want to overwrite
        # it
        if os.path.exists(out_dir):
            msg = out_dir+" already exists, overwrite?"
            if not input("%s (y/N) " % msg).lower() == 'y':
                quit("Bye.")
            if os.path.isdir(out_dir):
                shutil.rmtree(out_dir)
            else:
                os.remove(out_dir)
        self.out_dir = out_dir

        if version == "dev":
            self.version = {'major': 10000, 'minor': 0, 'revision': 0,
                            'num': version}
        else:
            major, minor, revision = version.split(".")
            if "-rc" in revision:
                revision, rc = revision.split("-rc")
            else:
                rc = -1
            self.version = {'major': int(major), 'minor': int(minor),
                            'revision': int(revision), 'rc': int(rc),
                            'num': version}

        # Initialise prov document
        self.doc = ProvDocument()
        self._add_namespaces()

        # A temp directory that will contain the exported data
        self.export_dir = tempfile.mkdtemp(prefix="nidm-", dir=out_path)

        self.prepend_path = ''

    def parse(self):
        """
        Parse a result directory to extract the pieces information to be
        stored in NIDM-Results.
        """

        try:
            # Methods: find_software, find_model_fitting, find_contrasts and
            # find_inferences should be defined in the children classes and
            # return a list of NIDM Objects as specified in the objects module

            # Object of type Software describing the neuroimaging software
            # package used for the analysis
            self.software = self._find_software()

            # List of objects of type ModelFitting describing the
            # model fitting step in NIDM-Results (main activity: Model
            # Parameters Estimation)
            self.model_fittings = self._find_model_fitting()

            # Dictionary of (key, value) pairs where where key is a tuple
            # containing the identifier of a ModelParametersEstimation object
            # and a tuple of identifiers of ParameterEstimateMap objects and
            # value is an object of type Contrast describing the contrast
            # estimation step in NIDM-Results (main activity: Contrast
            # Estimation)
            self.contrasts = self._find_contrasts()

            # Inference activity and entities
            # Dictionary of (key, value) pairs where key is the identifier of a
            # ContrastEstimation object and value is an object of type
            # Inference describing the inference step in NIDM-Results (main
            # activity: Inference)
            self.inferences = self._find_inferences()
        except Exception:
            self.cleanup()
            raise

    def cleanup(self,):
        if os.path.isdir(self.export_dir):
            shutil.rmtree(self.export_dir)

    def add_object(self, nidm_object, export_file=True):
        """
        Add a NIDMObject to a NIDM-Results export.
        """
        if not export_file:
            export_dir = None
        else:
            export_dir = self.export_dir

        if not isinstance(nidm_object, NIDMFile):
            nidm_object.export(self.version, export_dir)
        else:
            nidm_object.export(self.version, export_dir, self.prepend_path)
        # ProvDocument: add object to the bundle
        if nidm_object.prov_type == PROV['Activity']:
            self.bundle.activity(nidm_object.id,
                                 other_attributes=nidm_object.attributes)
        elif nidm_object.prov_type == PROV['Entity']:
            self.bundle.entity(nidm_object.id,
                               other_attributes=nidm_object.attributes)
        elif nidm_object.prov_type == PROV['Agent']:
            self.bundle.agent(nidm_object.id,
                              other_attributes=nidm_object.attributes)
        # self.bundle.update(nidm_object.p)

    def export(self):
        """
        Generate a NIDM-Results export.
        """
        try:
            if not os.path.isdir(self.export_dir):
                os.mkdir(self.export_dir)

            # Initialise main bundle
            self._create_bundle(self.version)

            self.add_object(self.software)

            # Add model fitting steps
            if not isinstance(self.model_fittings, list):
                self.model_fittings = list(self.model_fittings.values())

            for model_fitting in self.model_fittings:
                # Design Matrix
                # model_fitting.activity.used(model_fitting.design_matrix)
                self.bundle.used(model_fitting.activity.id,
                                 model_fitting.design_matrix.id)
                self.add_object(model_fitting.design_matrix)
                # *** Export visualisation of the design matrix
                self.add_object(model_fitting.design_matrix.image)

                if model_fitting.design_matrix.image.file is not None:
                    self.add_object(model_fitting.design_matrix.image.file)

                if model_fitting.design_matrix.hrf_models is not None:
                    # drift model
                    self.add_object(model_fitting.design_matrix.drift_model)

                if self.version['major'] > 1 or \
                        (self.version['major'] == 1 and
                         self.version['minor'] >= 3):
                    # Machine
                    # model_fitting.data.wasAttributedTo(model_fitting.machine)
                    self.bundle.wasAttributedTo(model_fitting.data.id,
                                                model_fitting.machine.id)
                    self.add_object(model_fitting.machine)

                    # Imaged subject or group(s)
                    for sub in model_fitting.subjects:
                        self.add_object(sub)
                        # model_fitting.data.wasAttributedTo(sub)
                        self.bundle.wasAttributedTo(model_fitting.data.id,
                                                    sub.id)

                # Data
                # model_fitting.activity.used(model_fitting.data)
                self.bundle.used(model_fitting.activity.id,
                                 model_fitting.data.id)
                self.add_object(model_fitting.data)

                # Error Model
                # model_fitting.activity.used(model_fitting.error_model)
                self.bundle.used(model_fitting.activity.id,
                                 model_fitting.error_model.id)
                self.add_object(model_fitting.error_model)

                # Parameter Estimate Maps
                for param_estimate in model_fitting.param_estimates:
                    # param_estimate.wasGeneratedBy(model_fitting.activity)
                    self.bundle.wasGeneratedBy(param_estimate.id,
                                               model_fitting.activity.id)
                    self.add_object(param_estimate)
                    self.add_object(param_estimate.coord_space)
                    self.add_object(param_estimate.file)

                    if param_estimate.derfrom is not None:
                        self.bundle.wasDerivedFrom(param_estimate.id,
                                                   param_estimate.derfrom.id)
                        self.add_object(param_estimate.derfrom)
                        self.add_object(param_estimate.derfrom.file,
                                        export_file=False)

                # Residual Mean Squares Map
                # model_fitting.rms_map.wasGeneratedBy(model_fitting.activity)
                self.add_object(model_fitting.rms_map)
                self.bundle.wasGeneratedBy(model_fitting.rms_map.id,
                                           model_fitting.activity.id)
                self.add_object(model_fitting.rms_map.coord_space)
                self.add_object(model_fitting.rms_map.file)
                if model_fitting.rms_map.derfrom is not None:
                    self.bundle.wasDerivedFrom(
                        model_fitting.rms_map.id,
                        model_fitting.rms_map.derfrom.id)
                    self.add_object(model_fitting.rms_map.derfrom)
                    self.add_object(model_fitting.rms_map.derfrom.file,
                                    export_file=False)

                # Resels per Voxel Map
                if model_fitting.rpv_map is not None:
                    self.add_object(model_fitting.rpv_map)
                    self.bundle.wasGeneratedBy(model_fitting.rpv_map.id,
                                               model_fitting.activity.id)
                    self.add_object(model_fitting.rpv_map.coord_space)
                    self.add_object(model_fitting.rpv_map.file)
                    if model_fitting.rpv_map.inf_id is not None:
                        self.bundle.used(model_fitting.rpv_map.inf_id,
                                         model_fitting.rpv_map.id)
                    if model_fitting.rpv_map.derfrom is not None:
                        self.bundle.wasDerivedFrom(
                            model_fitting.rpv_map.id,
                            model_fitting.rpv_map.derfrom.id)
                        self.add_object(model_fitting.rpv_map.derfrom)
                        self.add_object(model_fitting.rpv_map.derfrom.file,
                                        export_file=False)

                # Mask
                # model_fitting.mask_map.wasGeneratedBy(model_fitting.activity)
                self.bundle.wasGeneratedBy(model_fitting.mask_map.id,
                                           model_fitting.activity.id)
                self.add_object(model_fitting.mask_map)
                if model_fitting.mask_map.derfrom is not None:
                    self.bundle.wasDerivedFrom(
                        model_fitting.mask_map.id,
                        model_fitting.mask_map.derfrom.id)
                    self.add_object(model_fitting.mask_map.derfrom)
                    self.add_object(model_fitting.mask_map.derfrom.file,
                                    export_file=False)

                # Create coordinate space export
                self.add_object(model_fitting.mask_map.coord_space)
                # Create "Mask map" entity
                self.add_object(model_fitting.mask_map.file)

                # Grand Mean map
                # model_fitting.grand_mean_map.wasGeneratedBy(model_fitting.activity)
                self.bundle.wasGeneratedBy(model_fitting.grand_mean_map.id,
                                           model_fitting.activity.id)
                self.add_object(model_fitting.grand_mean_map)
                # Coordinate space entity
                self.add_object(model_fitting.grand_mean_map.coord_space)
                # Grand Mean Map entity
                self.add_object(model_fitting.grand_mean_map.file)

                # Model Parameters Estimation activity
                self.add_object(model_fitting.activity)
                self.bundle.wasAssociatedWith(model_fitting.activity.id,
                                              self.software.id)
                # model_fitting.activity.wasAssociatedWith(self.software)
                # self.add_object(model_fitting)

            # Add contrast estimation steps
            analysis_masks = dict()
            for (model_fitting_id, pe_ids), contrasts in list(
                    self.contrasts.items()):
                for contrast in contrasts:
                    model_fitting = self._get_model_fitting(model_fitting_id)
                    # for contrast in contrasts:
                    # contrast.estimation.used(model_fitting.rms_map)
                    self.bundle.used(contrast.estimation.id,
                                     model_fitting.rms_map.id)
                    # contrast.estimation.used(model_fitting.mask_map)
                    self.bundle.used(contrast.estimation.id,
                                     model_fitting.mask_map.id)
                    analysis_masks[contrast.estimation.id] = \
                        model_fitting.mask_map.id
                    self.bundle.used(contrast.estimation.id,
                                     contrast.weights.id)
                    self.bundle.used(contrast.estimation.id,
                                     model_fitting.design_matrix.id)
                    # contrast.estimation.wasAssociatedWith(self.software)
                    self.bundle.wasAssociatedWith(contrast.estimation.id,
                                                  self.software.id)

                    for pe_id in pe_ids:
                        # contrast.estimation.used(pe_id)
                        self.bundle.used(contrast.estimation.id, pe_id)

                    # Create estimation activity
                    self.add_object(contrast.estimation)

                    # Create contrast weights
                    self.add_object(contrast.weights)

                    if contrast.contrast_map is not None:
                        # Create contrast Map
                        # contrast.contrast_map.wasGeneratedBy(contrast.estimation)
                        self.bundle.wasGeneratedBy(contrast.contrast_map.id,
                                                   contrast.estimation.id)
                        self.add_object(contrast.contrast_map)
                        self.add_object(contrast.contrast_map.coord_space)
                        # Copy contrast map in export directory
                        self.add_object(contrast.contrast_map.file)

                        if contrast.contrast_map.derfrom is not None:
                            self.bundle.wasDerivedFrom(
                                contrast.contrast_map.id,
                                contrast.contrast_map.derfrom.id)
                            self.add_object(contrast.contrast_map.derfrom)
                            self.add_object(contrast.contrast_map.derfrom.file,
                                            export_file=False)

                    # Create Std Err. Map (T-tests) or Explained Mean Sq. Map
                    # (F-tests)
                    # contrast.stderr_or_expl_mean_sq_map.wasGeneratedBy
                    # (contrast.estimation)
                    stderr_explmeansq_map = (
                        contrast.stderr_or_expl_mean_sq_map)
                    self.bundle.wasGeneratedBy(
                        stderr_explmeansq_map.id,
                        contrast.estimation.id)
                    self.add_object(stderr_explmeansq_map)
                    self.add_object(
                        stderr_explmeansq_map.coord_space)
                    if isinstance(stderr_explmeansq_map,
                                  ContrastStdErrMap) and \
                            stderr_explmeansq_map.contrast_var:
                        self.add_object(
                            stderr_explmeansq_map.contrast_var)
                        if stderr_explmeansq_map.var_coord_space:
                            self.add_object(
                                stderr_explmeansq_map.var_coord_space)
                        if stderr_explmeansq_map.contrast_var.coord_space:
                            self.add_object(
                                stderr_explmeansq_map.contrast_var.coord_space)
                        self.add_object(
                            stderr_explmeansq_map.contrast_var.file,
                            export_file=False)
                        self.bundle.wasDerivedFrom(
                            stderr_explmeansq_map.id,
                            stderr_explmeansq_map.contrast_var.id)
                    self.add_object(stderr_explmeansq_map.file)

                    # Create Statistic Map
                    # contrast.stat_map.wasGeneratedBy(contrast.estimation)
                    self.bundle.wasGeneratedBy(contrast.stat_map.id,
                                               contrast.estimation.id)
                    self.add_object(contrast.stat_map)
                    self.add_object(contrast.stat_map.coord_space)
                    # Copy Statistical map in export directory
                    self.add_object(contrast.stat_map.file)

                    if contrast.stat_map.derfrom is not None:
                        self.bundle.wasDerivedFrom(
                            contrast.stat_map.id,
                            contrast.stat_map.derfrom.id)
                        self.add_object(contrast.stat_map.derfrom)
                        self.add_object(contrast.stat_map.derfrom.file,
                                        export_file=False)

                    # Create Z Statistic Map
                    if contrast.z_stat_map:
                        # contrast.z_stat_map.wasGeneratedBy(contrast.estimation)
                        self.bundle.wasGeneratedBy(contrast.z_stat_map.id,
                                                   contrast.estimation.id)
                        self.add_object(contrast.z_stat_map)
                        self.add_object(contrast.z_stat_map.coord_space)
                        # Copy Statistical map in export directory
                        self.add_object(contrast.z_stat_map.file)

                    # self.add_object(contrast)

            # Add inference steps
            for contrast_id, inferences in list(self.inferences.items()):
                contrast = self._get_contrast(contrast_id)

                for inference in inferences:
                    if contrast.z_stat_map:
                        used_id = contrast.z_stat_map.id
                    else:
                        used_id = contrast.stat_map.id
                    # inference.inference_act.used(used_id)
                    self.bundle.used(inference.inference_act.id, used_id)
                    # inference.inference_act.wasAssociatedWith(self.software)
                    self.bundle.wasAssociatedWith(inference.inference_act.id,
                                                  self.software.id)

                    # self.add_object(inference)
                    # Excursion set
                    # inference.excursion_set.wasGeneratedBy(inference.inference_act)
                    self.bundle.wasGeneratedBy(inference.excursion_set.id,
                                               inference.inference_act.id)
                    self.add_object(inference.excursion_set)
                    self.add_object(inference.excursion_set.coord_space)
                    if inference.excursion_set.visu is not None:
                        self.add_object(inference.excursion_set.visu)
                        if inference.excursion_set.visu.file is not None:
                            self.add_object(inference.excursion_set.visu.file)
                    # Copy "Excursion set map" file in export directory
                    self.add_object(inference.excursion_set.file)
                    if inference.excursion_set.clust_map is not None:
                        self.add_object(inference.excursion_set.clust_map)
                        self.add_object(inference.excursion_set.clust_map.file)
                        self.add_object(
                            inference.excursion_set.clust_map.coord_space)

                    if inference.excursion_set.mip is not None:
                        self.add_object(inference.excursion_set.mip)
                        self.add_object(inference.excursion_set.mip.file)

                    # Height threshold
                    if inference.height_thresh.equiv_thresh is not None:
                        for equiv in inference.height_thresh.equiv_thresh:
                            self.add_object(equiv)
                    self.add_object(inference.height_thresh)

                    # Extent threshold
                    if inference.extent_thresh.equiv_thresh is not None:
                        for equiv in inference.extent_thresh.equiv_thresh:
                            self.add_object(equiv)
                    self.add_object(inference.extent_thresh)

                    # Display Mask (potentially more than 1)
                    if inference.disp_mask:
                        for mask in inference.disp_mask:
                            # inference.inference_act.used(mask)
                            self.bundle.used(inference.inference_act.id,
                                             mask.id)
                            self.add_object(mask)
                            # Create coordinate space entity
                            self.add_object(mask.coord_space)
                            # Create "Display Mask Map" entity
                            self.add_object(mask.file)

                            if mask.derfrom is not None:
                                self.bundle.wasDerivedFrom(mask.id,
                                                           mask.derfrom.id)
                                self.add_object(mask.derfrom)
                                self.add_object(mask.derfrom.file,
                                                export_file=False)

                    # Search Space
                    self.bundle.wasGeneratedBy(inference.search_space.id,
                                               inference.inference_act.id)
                    # inference.search_space.wasGeneratedBy(inference.inference_act)
                    self.add_object(inference.search_space)
                    self.add_object(inference.search_space.coord_space)
                    # Copy "Mask map" in export directory
                    self.add_object(inference.search_space.file)

                    # Peak Definition
                    if inference.peak_criteria:
                        # inference.inference_act.used(inference.peak_criteria)
                        self.bundle.used(inference.inference_act.id,
                                         inference.peak_criteria.id)
                        self.add_object(inference.peak_criteria)

                    # Cluster Definition
                    if inference.cluster_criteria:
                        # inference.inference_act.used(inference.cluster_criteria)
                        self.bundle.used(inference.inference_act.id,
                                         inference.cluster_criteria.id)
                        self.add_object(inference.cluster_criteria)

                    if inference.clusters:
                        # Clusters and peaks
                        for cluster in inference.clusters:
                            # cluster.wasDerivedFrom(inference.excursion_set)
                            self.bundle.wasDerivedFrom(
                                cluster.id, inference.excursion_set.id)
                            self.add_object(cluster)
                            for peak in cluster.peaks:
                                self.bundle.wasDerivedFrom(peak.id, cluster.id)
                                self.add_object(peak)
                                self.add_object(peak.coordinate)

                            if cluster.cog is not None:
                                self.bundle.wasDerivedFrom(cluster.cog.id,
                                                           cluster.id)
                                self.add_object(cluster.cog)
                                self.add_object(cluster.cog.coordinate)

                    # Inference activity
                    # inference.inference_act.wasAssociatedWith(inference.software_id)
                    # inference.inference_act.used(inference.height_thresh)
                    self.bundle.used(inference.inference_act.id,
                                     inference.height_thresh.id)
                    # inference.inference_act.used(inference.extent_thresh)
                    self.bundle.used(inference.inference_act.id,
                                     inference.extent_thresh.id)
                    self.bundle.used(inference.inference_act.id,
                                     analysis_masks[contrast.estimation.id])
                    self.add_object(inference.inference_act)

            # Write-out prov file
            self.save_prov_to_files()

            return self.out_dir
        except Exception:
            self.cleanup()
            raise

    def _get_model_fitting(self, mf_id):
        """
        Retreive model fitting with identifier 'mf_id' from the list of model
        fitting objects stored in self.model_fitting
        """
        for model_fitting in self.model_fittings:
            if model_fitting.activity.id == mf_id:
                return model_fitting

        raise Exception("Model fitting activity with id: " + str(mf_id) +
                        " not found.")

    def _get_contrast(self, con_id):
        """
        Retreive contrast with identifier 'con_id' from the list of contrast
        objects stored in self.contrasts
        """
        for contrasts in list(self.contrasts.values()):
            for contrast in contrasts:
                if contrast.estimation.id == con_id:
                    return contrast
        raise Exception("Contrast activity with id: " + str(con_id) +
                        " not found.")

    def _add_namespaces(self):
        """
        Add namespaces to NIDM document.
        """
        self.doc.add_namespace(NIDM)
        self.doc.add_namespace(NIIRI)
        self.doc.add_namespace(CRYPTO)
        self.doc.add_namespace(DCT)
        self.doc.add_namespace(DC)
        self.doc.add_namespace(NFO)
        self.doc.add_namespace(OBO)
        self.doc.add_namespace(SCR)
        self.doc.add_namespace(NIF)

    def _create_bundle(self, version):
        """
        Initialise NIDM-Results bundle.
        """
        # *** Bundle entity
        if not hasattr(self, 'bundle_ent'):
            self.bundle_ent = NIDMResultsBundle(nidm_version=version['num'])

        self.bundle = ProvBundle(identifier=self.bundle_ent.id)

        self.bundle_ent.export(self.version, self.export_dir)

        # # provn export
        # self.bundle = ProvBundle(identifier=bundle_id)

        self.doc.entity(self.bundle_ent.id,
                        other_attributes=self.bundle_ent.attributes)

        # *** NIDM-Results Export Activity
        if version['num'] not in ["1.0.0", "1.1.0"]:
            if not hasattr(self, 'export_act'):
                self.export_act = NIDMResultsExport()
            self.export_act.export(self.version, self.export_dir)
            # self.doc.update(self.export_act.p)
            self.doc.activity(self.export_act.id,
                              other_attributes=self.export_act.attributes)

        # *** bundle was Generated by NIDM-Results Export Activity
        if not hasattr(self, 'export_time'):
            self.export_time = str(datetime.datetime.now().time())

        if version['num'] in ["1.0.0", "1.1.0"]:
            self.doc.wasGeneratedBy(entity=self.bundle_ent.id,
                                    time=self.export_time)
        else:
            # provn
            self.doc.wasGeneratedBy(
                entity=self.bundle_ent.id, activity=self.export_act.id,
                time=self.export_time)

        # *** NIDM-Results Exporter (Software Agent)
        if version['num'] not in ["1.0.0", "1.1.0"]:
            if not hasattr(self, 'exporter'):
                self.exporter = self._get_exporter()
            self.exporter.export(self.version, self.export_dir)
            # self.doc.update(self.exporter.p)
            self.doc.agent(self.exporter.id,
                           other_attributes=self.exporter.attributes)

            self.doc.wasAssociatedWith(self.export_act.id, self.exporter.id)

    def _get_model_parameters_estimations(self, error_model):
        """
        Infer model estimation method from the 'error_model'. Return an object
        of type ModelParametersEstimation.
        """
        if error_model.dependance == NIDM_INDEPEDENT_ERROR:
            if error_model.variance_homo:
                estimation_method = STATO_OLS
            else:
                estimation_method = STATO_WLS
        else:
            estimation_method = STATO_GLS

        mpe = ModelParametersEstimation(estimation_method, self.software.id)

        return mpe

    def use_prefixes(self, ttl):
        prefix_file = os.path.join(os.path.dirname(__file__), 'prefixes.csv')
        context = dict()
        with open(prefix_file, encoding="ascii") as csvfile:
            reader = csv.reader(csvfile)
            next(reader, None)  # skip the headers
            for alphanum_id, prefix, uri in reader:
                if alphanum_id in ttl:
                    context[prefix] = uri
                    ttl = "@prefix " + prefix + ": <" + uri + "> .\n" + ttl
                    ttl = ttl.replace(alphanum_id, prefix + ":")
                    if uri in ttl:
                        ttl = ttl.replace(alphanum_id, prefix + ":")
                elif uri in ttl:
                    context[prefix] = uri
                    ttl = "@prefix " + prefix + ": <" + uri + "> .\n" + ttl
                    ttl = ttl.replace(alphanum_id, prefix + ":")
        return (ttl, context)

    def save_prov_to_files(self, showattributes=False):
        """
        Write-out provn serialisation to nidm.provn.
        """
        self.doc.add_bundle(self.bundle)
        # provn_file = os.path.join(self.export_dir, 'nidm.provn')
        # provn_fid = open(provn_file, 'w')
        # # FIXME None
        # # provn_fid.write(self.doc.get_provn(4).replace("None", "-"))
        # provn_fid.close()

        ttl_file = os.path.join(self.export_dir, 'nidm.ttl')
        ttl_txt = self.doc.serialize(format='rdf', rdf_format='turtle')
        ttl_txt, json_context = self.use_prefixes(ttl_txt)

        # Add namespaces to json-ld context
        for namespace in self.doc._namespaces.get_registered_namespaces():
            json_context[namespace._prefix] = namespace._uri
        for namespace in \
                list(self.doc._namespaces._default_namespaces.values()):
            json_context[namespace._prefix] = namespace._uri
        json_context["xsd"] = "http://www.w3.org/2000/01/rdf-schema#"

        # Work-around to issue with INF value in rdflib (reported in
        # https://github.com/RDFLib/rdflib/pull/655)
        ttl_txt = ttl_txt.replace(' inf ', ' "INF"^^xsd:float ')
        with open(ttl_file, 'w') as ttl_fid:
            ttl_fid.write(ttl_txt)

        # print(json_context)
        jsonld_file = os.path.join(self.export_dir, 'nidm.json')
        jsonld_txt = self.doc.serialize(format='rdf', rdf_format='json-ld',
                                        context=json_context)
        with open(jsonld_file, 'w') as jsonld_fid:
            jsonld_fid.write(jsonld_txt)

        # provjsonld_file = os.path.join(self.export_dir, 'nidm.provjsonld')
        # provjsonld_txt = self.doc.serialize(format='jsonld')
        # with open(provjsonld_file, 'w') as provjsonld_fid:
        #     provjsonld_fid.write(provjsonld_txt)

        # provn_file = os.path.join(self.export_dir, 'nidm.provn')
        # provn_txt = self.doc.serialize(format='provn')
        # with open(provn_file, 'w') as provn_fid:
        #     provn_fid.write(provn_txt)

        # Post-processing
        if not self.zipped:
            # Just rename temp directory to output_path
            os.rename(self.export_dir, self.out_dir)
        else:
            # Create a zip file that contains the content of the temp directory
            os.chdir(self.export_dir)
            zf = zipfile.ZipFile(os.path.join("..", self.out_dir), mode='w')
            try:
                for root, dirnames, filenames in os.walk("."):
                    for filename in filenames:
                        zf.write(os.path.join(filename))
                shutil.rmtree(os.path.join("..", self.export_dir))
            finally:
                zf.close()
                os.chdir("..")

        # ttl_fid = open(ttl_file, 'w');
        # serialization is done in xlm rdf
        # graph = Graph()
        # graph.parse(data=self.doc.serialize(format='rdf'), format="xml")
        # ttl_fid.write(graph.serialize(format="turtle"))
        # ttl_fid.write(self.doc.serialize(format='rdf').
            # replace("inf", '"INF"'))
        # ttl_fid.close()
        # print("provconvert -infile " + provn_file + " -outfile " + ttl_file)
        # check_call("provconvert -infile " + provn_file +
        #            " -outfile " + ttl_file, shell=True)



"""
Generic objects supporting the classes defined in NIDM-Results.

Specification: http://nidm.nidash.org/specs/nidm-results.html

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2013-2014
"""
from prov.model import Identifier
from prov.identifier import QualifiedName
import numpy as np
import os
from nidmresults.objects.constants import *
import nibabel as nib
import shutil
import hashlib
import uuid
import json
from prov.model import Literal
from prov.constants import XSD_STRING
import warnings
import zipfile


class NIDMObject(object):

    """
    Generic class, parent of all objects describing a NIDM entity, activity
    or agent
    """

    def __init__(self, oid=None):
        if oid is None:
            self.id = NIIRI[str(uuid.uuid4())]
        else:
            if not type(oid) is QualifiedName:
                oid = NIIRI.qname(Identifier(oid))
            self.id = oid

    def __str__(self):
        value = ""
        if hasattr(self, 'value'):
            value = ": " + self.value
        location = ""
        if hasattr(self, 'file'):
            if hasattr(self.file, 'path'):
                location = " - " + self.file.path
        return '"' + self.label + '"' + value + location

    def __repr__(self):
        return '<"' + self.label + '" ' + \
               str(self.id).replace("niiri:", "").replace(NIIRI._uri, "")[0:8]\
               + '>'

    def add_attributes(self, attributes):
        if hasattr(self, 'attributes'):
            if isinstance(attributes, tuple):
                attributes = list(attributes)
            if isinstance(self.attributes, tuple):
                self.attributes = list(self.attributes)

            if isinstance(attributes, dict):
                attributes = [[k, v] for k, v in attributes.items()]
            if isinstance(self.attributes, dict):
                self.attributes = [[k, v] for k, v in self.attributes.items()]

            self.attributes = attributes + self.attributes
        else:
            self.attributes = attributes


class NIDMResultsBundle(NIDMObject):
    """
    Object representing a NIDM-Results bundle entity.
    """

    def __init__(self, nidm_version=None, label=None, oid=None):
        super(NIDMResultsBundle, self).__init__(oid=oid)
        self.type = NIDM_RESULTS
        self.nidm_version = nidm_version
        if label is None:
            self.label = "NIDM-Results"
        else:
            self.label = label
        self.prov_type = PROV['Bundle']

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix dctype: <http://purl.org/dc/dcmitype/>

SELECT * WHERE
{
    """ + oid_var + """ a nidm_NIDMResults: ;
    rdfs:label ?label ;
    nidm_version: ?nidm_version .
}
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entity.
        """
        self.add_attributes([
            (PROV['type'], self.type),
            (PROV['type'], PROV['Bundle']),
            # Explicitely add bundle type
            (PROV['label'], self.label),
            (NIDM_VERSION, self.nidm_version),
        ])


class CoordinateSpace(NIDMObject):
    """
    Object representing a CoordinateSpace entity.
    """

    def __init__(self, coordinate_system, nifti_file=None, vox_to_world=None,
                 vox_size=None, dimensions=None, numdim=None, units=None,
                 oid=None, label="Coordinate space"):
        super(CoordinateSpace, self).__init__(oid)
        
        if not isinstance(coordinate_system, QualifiedName):
            coordinate_system = NIDM.qname(coordinate_system)

        self.coordinate_system = coordinate_system
        self.type = NIDM_COORDINATE_SPACE
        self.prov_type = PROV['Entity']
        self.label = label

        if (vox_to_world is None) and (vox_size is None) and\
                (dimensions is None) and (numdim is None) and\
                (units is None) and \
                (nifti_file is not None):
            thresImg = nib.load(nifti_file)
            thresImgHdr = thresImg.get_header()

            numdim = len(thresImg.shape)
            dimensions = np.asarray(thresImg.shape)
            # FIXME: is vox_to_world the qform?
            vox_to_world = thresImg.get_qform()
            vox_size = thresImgHdr['pixdim'][1:(numdim + 1)]
            # FIXME: this gives mm, sec => what is wrong: FSL file, nibabel,
            # other?
            # units = str(thresImgHdr.get_xyzt_units()).strip('()')
            units = ["mm", "mm", "mm"]

        self.number_of_dimensions = numdim
        if not type(vox_to_world) is np.ndarray:
            # This is useful if info was read from a NIDM pack
            vox_to_world = np.array(json.loads(vox_to_world))
            dimensions = np.array(json.loads(dimensions))
            units = json.loads(units)
            vox_size = np.array(json.loads(vox_size))

        self.voxel_to_world = vox_to_world
        self.voxel_size = vox_size
        self.dimensions = dimensions
        self.units = units

    def is_mni(self):
        mni_coords = [
            NIDM_MNI_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_LINEAR_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR2009A_ASYMMETRIC_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR2009A_SYMMETRIC_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR2009B_ASYMMETRIC_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR2009B_SYMMETRIC_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR2009C_ASYMMETRIC_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR2009C_SYMMETRIC_COORDINATE_SYSTEM.uri,
            NIDM_ICBM_MNI152_NON_LINEAR6TH_GENERATION_COORDINATE_SYSTEM.uri,
            NIDM_ICBM452_AIR_COORDINATE_SYSTEM.uri,
            NIDM_ICBM452_WARP5_COORDINATE_SYSTEM.uri,
            NIDM_IXI549_COORDINATE_SYSTEM.uri,
            NIDM_MNI305_COORDINATE_SYSTEM.uri]

        if str(self.coordinate_system) in mni_coords:
            return True
        else:
            return False

    def is_talairach(self):
        if str(self.coordinate_system) in \
                [NIDM_TALAIRACH_COORDINATE_SYSTEM.uri]:
            return True
        else:
            return False

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_CoordinateSpace: <http://purl.org/nidash/nidm#NIDM_0000016>
prefix nidm_voxelToWorldMapping: <http://purl.org/nidash/nidm#NIDM_0000132>
prefix nidm_voxelUnits: <http://purl.org/nidash/nidm#NIDM_0000133>
prefix nidm_voxelSize: <http://purl.org/nidash/nidm#NIDM_0000131>
prefix nidm_inWorldCoordinateSystem: <http://purl.org/nidash/nidm#NIDM_0000105>
prefix nidm_MNICoordinateSystem: <http://purl.org/nidash/nidm#NIDM_0000051>
prefix nidm_numberOfDimensions: <http://purl.org/nidash/nidm#NIDM_0000112>
prefix nidm_dimensionsInVoxels: <http://purl.org/nidash/nidm#NIDM_0000090>


SELECT ?oid ?label ?vox_to_world ?units ?vox_size ?coordinate_system ?numdim
?dimensions
        WHERE
        {
    """ + oid_var + """ a nidm_CoordinateSpace: ;
    rdfs:label ?label ;
    nidm_voxelToWorldMapping: ?vox_to_world ;
    nidm_voxelUnits: ?units ;
    nidm_voxelSize: ?vox_size ;
    nidm_inWorldCoordinateSystem: ?coordinate_system ;
    nidm_numberOfDimensions: ?numdim ;
    nidm_dimensionsInVoxels: ?dimensions .
    }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        self.add_attributes({
            PROV['type']: self.type,
            NIDM_DIMENSIONS_IN_VOXELS: json.dumps(self.dimensions.tolist()),
            NIDM_NUMBER_OF_DIMENSIONS: self.number_of_dimensions,
            NIDM_VOXEL_TO_WORLD_MAPPING:
            json.dumps(self.voxel_to_world.tolist()),
            NIDM_IN_WORLD_COORDINATE_SYSTEM: self.coordinate_system,
            NIDM_VOXEL_UNITS: json.dumps(self.units),
            NIDM_VOXEL_SIZE: json.dumps(self.voxel_size.tolist()),
            PROV['label']: self.label})


class NIDMFile(NIDMObject):
    """
    Object representing a File (to be used as attribute of another class)
    """
    def __init__(self, rdf_id, location, filename=None,
                 sha=None, fmt=None, temporary=False):
        super(NIDMFile, self).__init__()
        self.prov_type = PROV['Entity']
        self.path = location
        if filename is None:
            # Keep same file name
            path, self.filename = os.path.split(self.path)
        else:
            self.filename = filename

        # NIDMFile is not a NIDM class defined in the owl file
        self.type = None
        self.id = rdf_id
        self.label = "'NIDM file'"  # used if display is called

        self.sha = sha
        self.fmt = fmt
        self.temporary = temporary

    def is_nifti(self):
        if self.path is not None:
            name = self.path
        else:
            name = self.filename

        return name.endswith(".nii") or \
            name.endswith(".nii.gz") or \
            name.endswith(".img") or \
            name.endswith(".hrd")

    def get_sha_sum(self, nifti_file):
        nifti_img = nib.load(nifti_file)
        data = nifti_img.get_data()
        # Fix needed as in https://github.com/pymc-devs/pymc/issues/327
        if not data.flags["C_CONTIGUOUS"]:
            data = np.ascontiguousarray(data)

        return hashlib.sha512(data).hexdigest()

    def export(self, nidm_version, export_dir, prepend_path):
        """
        Copy file over of export_dir and create corresponding triples
        """
        if self.path is not None:
            if export_dir is not None:
                # Copy file only if export_dir is not None
                new_file = os.path.join(export_dir, self.filename)
                if not self.path == new_file:
                    if prepend_path.endswith('.zip'):
                        with zipfile.ZipFile(prepend_path) as z:
                            extracted = z.extract(str(self.path), export_dir)
                            shutil.move(extracted, new_file)
                    else:
                        if prepend_path:
                            file_copied = os.path.join(prepend_path, self.path)
                        else:
                            file_copied = self.path
                        shutil.copy(file_copied, new_file)

                    if self.temporary:
                        os.remove(self.path)
            else:
                new_file = self.path

        if nidm_version['num'] in ["1.0.0", "1.1.0"]:
            loc = Identifier("file://./" + self.filename)
        else:
            loc = Identifier(self.filename)

        self.add_attributes([(NFO['fileName'], self.filename)])

        if export_dir:
            self.add_attributes([(PROV['atLocation'], loc)])

        if nidm_version['num'] in ("1.0.0", "1.1.0"):
            path, org_filename = os.path.split(self.path)
            if (org_filename is not self.filename) \
                    and (not self.temporary):
                self.add_attributes([(NFO['fileName'], org_filename)])

        if self.is_nifti():
            if self.sha is None:
                self.sha = self.get_sha_sum(new_file)
            if self.fmt is None:
                self.fmt = "image/nifti"

            self.add_attributes([
                (CRYPTO['sha512'], self.sha),
                (DCT['format'], self.fmt)
            ])


class Image(NIDMObject):

    """
    Object representing an Image entity.
    """

    def __init__(self, image_file, filename, fmt='png', oid=None):
        super(Image, self).__init__(oid=oid)
        self.type = DCTYPE['Image']
        self.prov_type = PROV['Entity']
        self.file = NIDMFile(self.id, image_file, filename)
        self.label = ""  # Enable printing

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
        prefix dctype: <http://purl.org/dc/dcmitype/>


        SELECT * WHERE
                {
            """ + oid_var + """ a dctype:Image ;
            prov:atLocation ?image_file ;
            nfo:fileName ?filename ;
            dct:format ?fmt .
            }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entity.
        """
        if self.file is not None:
            self.add_attributes([
                (PROV['type'], self.type),
                (DCT['format'], "image/png"),
            ])


class NeuroimagingSoftware(NIDMObject):
    """
    Class representing a NeuroimagingSoftware Agent.
    """

    def __init__(self, software_type, version, label=None, feat_version=None,
                 oid=None):
        super(NeuroimagingSoftware, self).__init__(oid=oid)
        self.version = version

        if isinstance(software_type, QualifiedName):
            self.type = software_type
        else:
            if software_type.startswith('http'):
                self.type = Identifier(software_type)
            elif software_type.lower() == "fsl":
                self.type = SCR_FSL
            else:
                warnings.warn('Unrecognised software: ' + str(software_type))
                self.name = str(software_type)
                self.type = None

        # FIXME: get label from owl!
        if self.type == SCR_FSL:
            self.name = "FSL"
        elif self.type == SCR_SPM:
            self.name = "SPM"
        else:
            warnings.warn('Unrecognised software: ' + str(software_type))
            self.name = str(software_type)

        if not label:
            self.label = self.name
        else:
            self.label = label
        self.prov_type = PROV['Agent']
        self.feat_version = feat_version

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_softwareVersion: <http://purl.org/nidash/nidm#NIDM_0000122>
prefix fsl_featVersion: <http://purl.org/nidash/fsl#FSL_0000005>
prefix nidm_ModelParametersEstimation: <http://purl.org/nidash/nidm#NIDM_00000\
56>

SELECT DISTINCT * WHERE
        {
    """ + oid_var + """ a prov:SoftwareAgent ;
        nidm_softwareVersion: ?version .

    [] a nidm_ModelParametersEstimation: ;
        prov:wasAssociatedWith """ + oid_var + """ .

    OPTIONAL {""" + oid_var + """ a ?software_type .} .
    OPTIONAL {""" + oid_var + """ fsl_featVersion: ?feat_version .} .

    FILTER ( ?software_type NOT IN (prov:SoftwareAgent, prov:Agent) )
    }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        if nidm_version['major'] < 1 or \
                (nidm_version['major'] == 1 and nidm_version['minor'] < 3):
            self.type = NLX_OLD_FSL

        atts = (
            (PROV['type'], self.type),
            (PROV['type'], PROV['SoftwareAgent']),
            (PROV['label'], Literal(self.label, datatype=XSD_STRING)),
            (NIDM_SOFTWARE_VERSION, self.version)
            )

        if self.feat_version:
            atts = atts + ((FSL_FEAT_VERSION, self.feat_version),)

        self.add_attributes(atts)


class ExporterSoftware(NIDMObject):
    """
    Class representing a Software Agent.
    """

    def __init__(self, software_type, version, oid=None, label=None):
        super(ExporterSoftware, self).__init__(oid=oid)
        self.type = software_type
        self.prov_type = PROV['Agent']
        self.version = version

        if label is None:
            if software_type == NIDM_FSL:
                self.label = "nidmfsl"
            else:
                self.label = str(software_type)
        else:
            self.label = label

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_softwareVersion: <http://purl.org/nidash/nidm#NIDM_0000122>
prefix nidm_NIDMResultsExport: <http://purl.org/nidash/nidm#NIDM_0000166>

SELECT DISTINCT * WHERE
        {
    """ + oid_var + """ a prov:SoftwareAgent ;
        rdfs:label  ?label ;
        rdf:type ?software_type ;
        nidm_softwareVersion: ?version .

    FILTER ( ?software_type NOT IN (prov:SoftwareAgent, prov:Agent) )
    }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        self.add_attributes((
            (PROV['type'], self.type),
            (PROV['type'], PROV['SoftwareAgent']),
            (PROV['label'], self.label),
            (NIDM_SOFTWARE_VERSION, self.version))
        )


class NIDMResultsExport(NIDMObject):
    """
    Class representing a NIDM-Results Export activity.
    """
    def __init__(self, oid=None, label=None):
        super(NIDMResultsExport, self).__init__(oid=oid)
        self.type = NIDM_NIDM_RESULTS_EXPORT
        self.prov_type = PROV['Activity']
        if label is None:
            self.label = "NIDM-Results export"
        else:
            self.label = label

    @classmethod
    def get_query(klass, oid=None):
        if oid is None:
            oid_var = "?oid"
        else:
            oid_var = "<" + str(oid) + ">"

        query = """
prefix nidm_NIDMResultsExport: <http://purl.org/nidash/nidm#NIDM_0000166>

SELECT DISTINCT * WHERE
    {
    """ + oid_var + """ a nidm_NIDMResultsExport: ;
        rdfs:label ?label .
    }
        """
        return query

    def export(self, nidm_version, export_dir):
        """
        Create prov entities and activities.
        """
        self.add_attributes([
            (PROV['label'], self.label),
            (PROV['type'], self.type)])



""" Load and save NIDM-Results objects """
from .graph import NIDMResults

import os


def load(filename, to_replace=dict()):
    ''' Load NIDM-Results file given filename, guessing if it is a
    NIDM-Results pack or a JSON file

    Parameters
    ----------
    filename : string
       specification of file to load
    Returns
    -------
    nidmres : ``NIDMResults``
       NIDM-Results object
    '''
    if not os.path.exists(filename):
        raise IOException('File does not exist: %s' % filename)

    if filename.endswith('.json'):
        raise Exception('Minimal json file: not handled yet')
    elif filename.endswith('.nidm.zip'):
        nidm = NIDMResults.load_from_pack(filename, to_replace=to_replace)

    return nidm
